"use strict";

Object.defineProperty(exports, "__esModule", {
    value: true
});

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

var _babelTypes = require("babel-types");

var t = _interopRequireWildcard(_babelTypes);

var _babelTemplate = require("babel-template");

var _babelTemplate2 = _interopRequireDefault(_babelTemplate);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

/**
 * Template for import statements
 */
var compileImportTemplate = (0, _babelTemplate2.default)("\n    import * as IMPORT_NAME from 'SOURCE';\n", { sourceType: "module" });

/**
 * Template for "Register.customElement"
 */
var compileRegisterTemplate = (0, _babelTemplate2.default)("\n    REGISTER_VALUE.REGISTER_VALUE_ORIG.customElement(\n        CLASS_NAME, STORAGE_VALUE.STORAGE_VALUE_ORIG.get(CLASS_NAME).get(STORAGE_POINTER)\n    );\n");

/**
 * buildImportAst
 * @param data {{IMPORT_NAME: string, SOURCE: string}}
 * @returns {object} ast
 */
var buildImportAst = function buildImportAst(data) {
    var IMPORT_NAME = data.IMPORT_NAME,
        SOURCE = data.SOURCE;

    var ast = compileImportTemplate({
        IMPORT_NAME: t.identifier(IMPORT_NAME)
    });
    ast.source.value = SOURCE;
    ast._$importName = IMPORT_NAME;
    ast._$source = SOURCE;

    return ast;
};

/**
 * buildRegisterAst
 * @param data {{
     *     CLASS_NAME: string,
     *     REGISTER_VALUE: string,
     *     REGISTER_VALUE_ORIG: string,
     *     STORAGE_VALUE: string,
     *     STORAGE_VALUE_ORIG: string
     * }}
 * @returns {object}
 */
var buildRegisterAst = function buildRegisterAst(data) {

    return compileRegisterTemplate({
        CLASS_NAME: t.identifier(data.CLASS_NAME),
        STORAGE_POINTER: t.stringLiteral(data.STORAGE_POINTER),
        REGISTER_VALUE: t.identifier(data.REGISTER_VALUE),
        REGISTER_VALUE_ORIG: t.identifier(data.REGISTER_VALUE_ORIG),
        STORAGE_VALUE: t.identifier(data.STORAGE_VALUE),
        STORAGE_VALUE_ORIG: t.identifier(data.STORAGE_VALUE_ORIG)
    });
};

/**
 * getDecorator
 * @param decoratorName {string}
 * @returns {*}
 */
var getDecorator = function getDecorator(decoratorName) {

    if (!this.node.decorators) {
        return [];
    }

    return this.node.decorators.filter(function (deco) {
        var name = deco.expression.callee.name;

        if (name === decoratorName) {
            return true;
        }
    });
};

/**
 * addImports
 * @param path {object}
 * @param imports {array}
 * @param programImports {array}
 * @param programBody {object}
 */
var addImports = function addImports(path, imports, programImports, programBody) {
    var _this = this;

    imports.forEach(function (imp) {

        var result = null;
        var _iteratorNormalCompletion = true;
        var _didIteratorError = false;
        var _iteratorError = undefined;

        try {
            for (var _iterator = programImports[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
                var prImp = _step.value;


                imp.IMPORTED = imp.IMPORT_NAME === prImp.IMPORT_NAME && imp.SOURCE === prImp.SOURCE;

                if (imp.IMPORTED) {
                    result = prImp;
                    break;
                }
            }
        } catch (err) {
            _didIteratorError = true;
            _iteratorError = err;
        } finally {
            try {
                if (!_iteratorNormalCompletion && _iterator.return) {
                    _iterator.return();
                }
            } finally {
                if (_didIteratorError) {
                    throw _iteratorError;
                }
            }
        }

        if (!result) {
            var createdImportAst = buildImport.call(_this, imp, path);
            programBody.unshift(createdImportAst);
        }
    });
};

/**
 * buildImport
 * @param data {object}
 * @param path {object}
 * @returns {Object}
 */
var buildImport = function buildImport(data, path) {
    var IMPORT_NAME = data.IMPORT_NAME,
        SOURCE = data.SOURCE;

    var IMPORT_NAME_UID = path.scope.generateUidIdentifier(IMPORT_NAME).name;

    this.cache.set(IMPORT_NAME + "-value", IMPORT_NAME_UID);
    this.cache.set(IMPORT_NAME + "-value-orig", IMPORT_NAME);

    return buildImportAst({
        IMPORT_NAME: IMPORT_NAME_UID,
        SOURCE: SOURCE
    });
};

var setDecoratorState = function setDecoratorState(self, path) {

    var component = getDecorator.call(path, 'component');
    if (component.length) {
        self.cache.set('decorator', true);
    }
};

/**
 * getImportDeclaration
 * @param path {object}
 * @returns {Array}
 */
var getImportDeclaration = function getImportDeclaration(path) {

    // lookup for import
    var importPath = path.node.source.value;
    if (!importPath) {
        return;
    }

    // create import object { IMPORT_NAME: 'foo', SOURCE: './a' }
    var imports = path.node.specifiers.map(function (_ref) {
        var local = _ref.local;

        return {
            IMPORT_NAME: local.name,
            SOURCE: importPath
        };
    });

    return imports;
};

/**
 * getClassName
 * @param path
 * @returns {string};
 */
var getClassName = function getClassName(path) {
    return path.node.id.name;
};

function plugin() {

    return {
        pre: function pre() {
            this.cache = new Map();
            this.cache.set('decorator', false);
        },
        post: function post() {
            this.cache.clear();
        },

        visitor: {
            Program: function Program(path, state) {

                var optionsImports = state.opts.imports;
                var programBody = path.node.body;

                if (optionsImports.length !== 2) {
                    throw new Error('Please pass Register and storage');
                }

                /**
                 * Check that @component exist
                 */
                var self = this;
                path.traverse({
                    ClassDeclaration: function ClassDeclaration(path) {
                        return setDecoratorState(self, path);
                    }
                });
                if (!self.cache.get('decorator')) {
                    return;
                }

                /**
                 * Check that required imports already imported
                 */
                var programImports = [];
                path.traverse({
                    ImportDeclaration: function ImportDeclaration(path) {
                        var imports = getImportDeclaration(path);
                        programImports.push.apply(programImports, _toConsumableArray(imports));
                    }
                });

                addImports.call(this, path, optionsImports, programImports, programBody);
            },
            ClassDeclaration: function ClassDeclaration(path, state) {
                var _context;

                var storage_pointer = state.opts.storage_pointer;


                var component = (_context = path, getDecorator).call(_context, 'component');
                if (!component.length) {
                    return;
                }

                var _state$opts$imports = _slicedToArray(state.opts.imports, 2),
                    imp1 = _state$opts$imports[0],
                    imp2 = _state$opts$imports[1];

                if (imp1.IMPORTED && imp2.IMPORTED) {
                    return;
                }

                // build Register ast
                var registerAst = buildRegisterAst({
                    CLASS_NAME: getClassName(path),
                    STORAGE_POINTER: storage_pointer,
                    REGISTER_VALUE: this.cache.get('Register-value'),
                    REGISTER_VALUE_ORIG: this.cache.get('Register-value-orig'),
                    STORAGE_VALUE: this.cache.get('storage-value'),
                    STORAGE_VALUE_ORIG: this.cache.get('storage-value-orig')
                });

                if (path.parentPath && path.parentPath.type === "ExportNamedDeclaration") {
                    path = path.parentPath;
                }
                path.insertAfter(registerAst);
            }
        }
    };
}

exports.default = plugin;